import fnmatch
import rasterio as rio
from io import BytesIO
import numpy as np
from rasterio.io import MemoryFile
import pandas as pd

from .minio import get_client
from .BaseStorage import BaseStorage


class CloudStorage(BaseStorage):
    # store in s3
    def __init__(self, bucket="spai"):
        super().__init__()
        self.client = get_client()
        self.bucket = bucket
        if not self.client.bucket_exists(bucket):
            self.client.make_bucket(bucket)
            print(f"{bucket} created")

    def get_url(self, name):
        return self.client.presigned_get_object(self.bucket, name)

    def create_from_path(self, data, name):
        self.client.fput_object(self.bucket, name, data)
        return self.get_url(name)

    def create_from_image(self, data, name):
        image_bytes = BytesIO()
        # get format from file extension
        format = name.split(".")[-1]
        data.save(image_bytes, format=format)
        image_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, image_bytes, len(image_bytes.getvalue())
        )
        return self.get_url(name)

    def create_from_rasterio(self, x, name, ds, window=None):
        transform = ds.transform if window is None else ds.window_transform(window)
        with MemoryFile() as memfile:
            with memfile.open(
                driver="GTiff",
                count=1 if x.ndim < 3 else x.shape[0],
                height=x.shape[0] if x.ndim < 3 else x.shape[1],
                width=x.shape[1] if x.ndim < 3 else x.shape[2],
                dtype=np.uint8 if x.dtype == "bool" else x.dtype,
                crs=ds.crs,
                transform=transform,
                # nbits=1 if x.dtype == 'bool' else
            ) as dest_ds:
                bands = 1 if x.ndim < 3 else [i + 1 for i in range(x.shape[0])]
                dest_ds.write(x, bands)
            self.client.put_object(
                self.bucket, name, memfile, length=-1, part_size=10 * 1024 * 1024
            )
        return self.get_url(name)

    def create_from_array(self, data, name):
        array_bytes = BytesIO()
        np.save(array_bytes, data)
        array_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, array_bytes, len(array_bytes.getvalue())
        )
        return self.get_url(name)

    def create_from_csv(self, data, name):
        csv_bytes = BytesIO()
        data.to_csv(csv_bytes)
        csv_bytes.seek(0)
        self.client.put_object(self.bucket, name, csv_bytes, len(csv_bytes.getvalue()))
        return self.get_url(name)

    def create_from_json(self, data, name):
        json_bytes = BytesIO()
        data.to_json(json_bytes)
        json_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, json_bytes, len(json_bytes.getvalue())
        )
        return self.get_url(name)

    def list(self, pattern="*"):
        return fnmatch.filter(  # need to test
            [
                obj.object_name
                for obj in self.client.list_objects(self.bucket, recursive=True)
            ],
            pattern,
        )

    def read_object(self, name):
        return BytesIO(self.client.get_object(self.bucket, name).read())

    def read_from_array(self, name):
        return np.load(self.read_object(name))

    def read_from_rasterio(self, name):
        return rio.open(self.read_object(name))

    def read_from_csv(self, name):
        return pd.read_csv(self.read_object(name), index_col=0)

    def read_from_json(self, name):
        return pd.read_json(self.read_object(name))



from datetime import datetime, timedelta
from dotenv import load_dotenv

from spai.storage import Storage
from spai.data.satellite import explore_satellite_images, download_satellite_image

load_dotenv()

# create time range (3 months)
now = datetime.now()
last_months = now - timedelta(days=30 * 3)
time_interval = (last_months.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d"))

# explore available images
print("Looking for images in the range", time_interval)
images = explore_satellite_images(time_interval, cloud_cover=10)
if len(images) == 0:
    raise ValueError("No images found")

# download images and save locally
storage = Storage("data")
sensor = "S2L2A"  # or 'S2L1C'
existing_images = storage.list(f"{sensor}*.tif")
dates = [image.split("_")[1].split(".")[0] for image in existing_images]
print("Found", len(images), f"image{'s' if len(images) > 1 else ''}")
for image in images:
    # check if image is already downloaded
    date = image["date"].split("T")[0]
    if date in dates:
        print("Image already downloaded:", date)
        continue
    print("Downloading new image:", date)
    path = download_satellite_image(date, sensor, storage)
    print("Image saved at", path)



import numpy as np
from PIL import Image
import io
import rasterio


def ready_image(data, scale=1.0):
    # define mode
    if data.ndim == 3:
        if data.shape[-1] == 3:  # encode RGB image
            mode = "RGB"
        elif data.shape[-1] == 4:  # encode RGBA image
            mode = "RGBA"
        else:
            raise ValueError("Input array with 3 dimensions must be RGB or RGBA")
    elif data.ndim == 2:  # encode paletted image
        mode = "L"
    else:
        raise ValueError("Input array must have 2 or 3 dimensions")
    # add transparency
    transparency = (0, 0, 0)
    if isinstance(data, np.ma.MaskedArray):
        data = data.filled(0)
    # create image
    img = Image.fromarray(data, mode=mode)
    # rescale image if necessary
    if scale != 1.0:
        h, w = data.shape[:2]
        nh, nw = int(h / scale), int(w / scale)
        img.thumbnail([nw, nh], Image.ANTIALIAS)
    # make buffer
    buf = io.BytesIO()
    if data.ndim == 3 and data.shape[-1] == 4:
        img.save(buf, "png")
    else:
        img.save(buf, "png", transparency=transparency)
    buf.seek(0)
    return buf


def contrast_stretch(data, in_range, out_range, clip):
    lower_bound_in, upper_bound_in = in_range
    lower_bound_out, upper_bound_out = out_range

    out_data = data.astype("float64", copy=True)
    out_data -= lower_bound_in
    norm = upper_bound_in - lower_bound_in
    if abs(norm) > 1e-8:  # prevent division by 0
        out_data *= (upper_bound_out - lower_bound_out) / norm
    out_data += lower_bound_out
    if clip:
        np.clip(out_data, lower_bound_out, upper_bound_out, out=out_data)
    return out_data


def to_uint8(data, lower_bound, upper_bound):
    rescaled = contrast_stretch(data, (lower_bound, upper_bound), (0, 255), clip=True)
    return rescaled.astype(np.uint8)



from .SHS2Downloader import SHS2Downloader
from sentinelhub import DataCollection


class SHS2L1CDownloader(SHS2Downloader):
    def __init__(self, download_folder="/tmp"):
        super().__init__(download_folder)
        self.data_collection = DataCollection.SENTINEL2_L1C



import typer
from pathlib import Path
import os
import shutil
import sys
import requests

# Add the cli directory to the Python path
spai_cli_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(spai_cli_dir))

from cli.commands import hello
from cli import (
    load_and_validate_config,
    deploy_and_run_cloud,
    run_local,
)

app = typer.Typer()
app.add_typer(hello.app, name="hello")


@app.command()
def init():
    # ask for project name
    project = typer.prompt("Project name")
    # copy template
    template = Path(__file__).parent / "cli" / "project-template"
    shutil.copytree(template, Path(project))
    # change name to project in spai.config.yaml
    config = Path(project) / "spai.config.yml"
    os.system(f"sed -i 's/project-template/{project}/g' {config}")
    return typer.echo(f"Project {project} created")


@app.command()
def run(
    dir: Path = typer.Option(Path("."), "-d"),
    cloud: bool = typer.Option(False, "-c", "--cloud"),
):
    # make sure dir is absolute
    dir = Path(dir).resolve()
    # load and validate config
    config = load_and_validate_config(dir, typer, cloud)
    # print(config)
    # run project
    if cloud:
        return deploy_and_run_cloud(dir, config, typer)
    return run_local(dir, config, typer)


# API_URL = "http://localhost:8000"
API_URL = "http://148.113.137.45:8010"


@app.command()
def list(
    project: str = typer.Option(None, "-p", "--project"),
):
    project = "asd"
    response = requests.get(f"{API_URL}/projects/{project}")
    if response.status_code == 200:
        data = response.json()
        scripts = [k for k in data.keys()]
        if len(scripts) == 0:
            return typer.echo(f"No scripts running in project '{project}'.")
        return typer.echo(f"Scripts in project '{project}': {scripts}")
    return typer.echo(response.json()["detail"])


@app.command()
def stop(
    project: str = typer.Option(None, "-p", "--project"),
):
    # list services
    project = "asd"
    response = requests.get(f"{API_URL}/projects/{project}")
    data = response.json()
    scripts = [k for k in data.keys()]
    if len(scripts) == 0:
        return typer.echo(f"No scripts running in project '{project}'.")
    # delete services
    for script in scripts:
        typer.echo(f"Stopping script '{script}'...")
        response = requests.get(f"{API_URL}/stop/script/{script}")
        if response.status_code == 200:
            typer.echo(f"Stopped.")
        else:
            typer.echo("Something went wrong.")
            typer.echo(response.json()["detail"])
    return typer.echo(f"Stopped all scripts in project '{project}'.")


@app.command()
def logs(
    script: str,
):
    response = requests.get(f"{API_URL}/logs/script/{script}")
    if response.status_code == 200:
        typer.echo(response.json())
    else:
        typer.echo(response.json()["detail"])


if __name__ == "__main__":
    app()



import os
import requests

# API_URL = "http://localhost:8000"
API_URL = "http://148.113.137.45:8010"


def call_api(script_dir, endpoint, data):
    return requests.post(
        f"{API_URL}/{endpoint}/script",
        files={
            "script": open(script_dir / "main.py", "rb"),
            "requirements": open(script_dir / "requirements.txt", "rb")
            if "requirements.txt" in os.listdir(script_dir)
            else None,
            "env": open(script_dir / ".env", "rb")
            if ".env" in os.listdir(script_dir)
            else None,
        },
        data=data,
    )


def deploy_and_run_cloud(dir, config, typer):
    typer.echo(f"Deploying...")
    if config.scripts:
        typer.echo(f"Deploying scripts...")
        for script in config.scripts:
            # check if script is already deployed
            response = requests.get(f"{API_URL}/projects/{config.project}")
            scripts = list(response.json().keys())
            if script.name in scripts:
                typer.echo(f"Script '{script.name}' already deployed.")
                typer.echo(f"Stopping ...")  # confirm
                response = requests.get(f"{API_URL}/stop/script/{script.name}")
                if response.status_code == 200:
                    typer.echo(f"Stopped.")
                else:
                    typer.echo("Something went wrong.")
                    typer.echo(response.json()["detail"])
            # deploy script
            script_dir = dir / "scripts" / script.name
            if script.run_on_start:
                typer.echo(f"Running script '{script.name}'...")
                response = call_api(
                    script_dir, "run", {"name": script.name, "project": config.project}
                )
                if response.status_code == 200:
                    print(response.json())
                else:
                    typer.echo("Something went wrong.")
                    typer.echo(response.json()["detail"])
            if script.run_every:
                typer.echo(f"Setting up cronjob for script '{script.name}'...")
                response = call_api(
                    script_dir,
                    "schedule",
                    {"name": script.name, "run_every": script.run_every},
                )
                if response.status_code == 200:
                    print(response.json())
                else:
                    typer.echo("Something went wrong.")
                    typer.echo(response.json()["detail"])
